// -------------------------------------------------------------------------------- 
//!  @file  main.c                                                                  
//!  メイン処理のソースファイル                                                     
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <pico/stdlib.h>
#include "config.h"
#include "libvga/vga/vga.h"
#include "libvga/pad/pad.h"
#include "data.h"
#include "sound.h"


	enum {
		FP_ONE_						=	8,							//!< 固定小数点１ビット数
		FP_ONE						=	1 << FP_ONE_				//!< 固定小数点１
	} ;

	//! VRAM
	typedef struct {
//#		uint32_t					pattern[VGA_TILE_PATTERN_MAX][VGA_TILE_IMAGE_WIDTH*VGA_TILE_IMAGE_HEIGHT/8] ;	//!< パターンデータ
//#		uint16_t					palette[VGA_TILE_PALETTE_MAX][VGA_TILE_PALETTE_COUNT] ;							//!< パレットデータ

		uint16_t					colorbar[48][VGA_BMPH_WIDTH] ;													//!< カラーバー

		union {
			uint8_t					bmp8[VGA_BMP_HEIGHT][VGA_BMPH_WIDTH] ;											//!<  8bit ビットマップデータ
			uint16_t				bmp16[VGA_BMP_HEIGHT][VGA_BMPH_WIDTH] ;											//!< 16bit ビットマップデータ
			uint16_t				tile[VGA_TILE_HEIGHT_COUNT][VGA_TILE_WIDTH_COUNT] ;								//!< タイルデータ
		} bg ;																										//!< 背景データ

		struct {
			VGASPR					spr[VGA_SPR_MAX] ;																//!< スプライトデータ
			uint16_t				txt[VGA_TXT_HEIGHT_COUNT][VGA_TXT_WIDTH_COUNT] ;								//!< テキストデータ
		} fg ;																										//!< 前景データ
	} VRAM ;


static int16_t						m_sinData[256] ;				//!< サインデータ
static VRAM							m_vram ;						//!< VRAM
static uint8_t						m_spritePos ;					//!< スプライト位置
static uint8_t						m_bitmapPos ;					//!< ビットマップ位置


// -------------------------------------------------------------------------------- 
//!  16bit ビットマップ VRAM へテストデータを設定                                   
//!                                                                                 
//!  @return  なし                                                                  
// -------------------------------------------------------------------------------- 
static void setBitmap16( void )
{
	int		x, y, r, g, b, xm, ym ;

	// カラーバー
	for(y=0; y<16; y++) {
		for(x=0; x<VGA_BMPH_WIDTH; x++) {
			m_vram.colorbar[y+ 0][x] = ((x * 16) / VGA_BMPH_WIDTH) << 8 ;
			m_vram.colorbar[y+16][x] = ((x * 16) / VGA_BMPH_WIDTH) << 4 ;
			m_vram.colorbar[y+32][x] = ((x * 16) / VGA_BMPH_WIDTH) << 0 ;
		}
	}

	// グラデーション
	for(y=0; y<VGA_BMP_HEIGHT; y++) {
		for(x=0; x<VGA_BMPH_WIDTH; x++) {
			xm = (x * 16) / VGA_BMPH_WIDTH ;
			ym = (y * 16) / VGA_BMP_HEIGHT ;

			r = xm ;
			g = ym ;
			b = ((15 - xm) * (15 - ym)) / 15 ;

			m_vram.bg.bmp16[y][x] = (uint16_t )((r << 0) | (g << 4) | (b << 8)) ;
		}
	}
}


// -------------------------------------------------------------------------------- 
//!  スプライト VRAM へテストデータを設定                                           
//!                                                                                 
//!  @return  なし                                                                  
// -------------------------------------------------------------------------------- 
static void setSpriteData( void )
{
	VGASPR	*p_spr ;
	int		index, x, y ;


	p_spr = &m_vram.fg.spr[0] ;

	// 表示
	for(y=0; y<3; y++) {
		for(x=0; x<3*5; x++,p_spr++) {
			index       = (x / 3) & 1 ;
			p_spr->x    = (32 + (x * 16)) | (index << VGASPR_X_OFS_BIT) ;
			p_spr->y    =       (y * 16)  |           VGASPR_Y_OFS ;
			p_spr->attr = (index<<VGASPR_ATTR_PC_BIT) | VGASPR_ATTR_PR | ((x % 3) + ((y % 3) * 3)) ;
		}
	}

	// 非表示
	for(; index<VGA_SPR_MAX; index++,p_spr++) {
		p_spr->y = VGA_IMAGE_HEIGHT ;
	}
}


// -------------------------------------------------------------------------------- 
//!  テキスト VRAM へテストデータを設定                                             
//!                                                                                 
//!  @return  なし                                                                  
// -------------------------------------------------------------------------------- 
static void setTextData( void )
{
	uint16_t	attr ;
	int			x, y ;


	attr = 0 ;

	for(y=0; y<4; y++,attr++) {
		for(x=0; x<VGA_TXT_WIDTH_COUNT; x++) {
			if( x < 20 ) {
				m_vram.fg.txt[y][x] = (1 << VGATXT_PC_BIT) | (attr << VGATXT_OP_BIT) | (0x80 + x) ;
			} else {
				m_vram.fg.txt[y][x] = VGATXT_OP | ' ' ;
			}
		}
	}
}


// -------------------------------------------------------------------------------- 
//!  テキスト VRAM へ文字列を書き込む                                               
//!                                                                                 
//!  @param [in] x         開始ｘ座標                                               
//!  @param [in] y         開始ｙ座標                                               
//!  @param [in] p_string  文字列アドレス                                           
//!  @param [in] pal       パレット番号                                             
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
static void writeString(int x, int y, const char *p_string, int pal)
{
	uint16_t	*p_txt, attr ;


	p_txt = &m_vram.fg.txt[y][x] ;
	attr  = (pal << VGATXT_PC_BIT) | VGATXT_OP ;

	while( *p_string != '\0' ) {
		*p_txt++ = attr | *p_string++ ;
	}
}


// -------------------------------------------------------------------------------- 
//!  VSync 割り込みサービスルーチン                                                 
//!                                                                                 
//!  @param [in] gpio    GPIO 番号                                                  
//!  @param [in] events  イベント                                                   
//!  @return             なし                                                       
// -------------------------------------------------------------------------------- 
static void isr_vsync(uint gpio, uint32_t events)
{
	sound_vsync() ;

	m_spritePos++ ;

	vga_setFgSprite(&m_vram.fg.spr[0], &g_sprPatData[0][0], &g_palData[0][0]) ;
	vga_setBgBitmap320x240x16( (const uint32_t *)&m_vram.colorbar[0][0] ) ;
	vga_setBitmapOffset(0, 0) ;

	if( ++m_bitmapPos == 240 ) {
		m_bitmapPos = 0 ;
	}
}


// -------------------------------------------------------------------------------- 
//!  HSync 処理                                                                     
//!                                                                                 
//!  @param [in] line  行                                                           
//!  @return           なし                                                         
// -------------------------------------------------------------------------------- 
static void RAMFUNC hsync( uint32_t line )
{
	if( line == 48 ) {
		vga_setBgBitmap320x240x16( (const uint32_t *)&m_vram.bg.bmp16[0][0] ) ;
		vga_setBitmapOffset(0, m_bitmapPos) ;
	}

	if( (uint )(line-152) < 32 ) {
		vga_setFgText(&m_vram.fg.txt[0][0], &g_txtPatData[0][0], &g_palData[0][0]) ;
		vga_setTextOffset(0, 240-152) ;
	} else {
		vga_setFgSprite(&m_vram.fg.spr[0], &g_sprPatData[0][0], &g_palData[0][0]) ;
		vga_setSpriteOffset((int32_t )-m_sinData[(m_spritePos+line)&255]*96/256, line-(line%48)) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  メイン処理                                                                     
//!                                                                                 
//!  @param [in] argc  引数の数                                                     
//!  @param [in] argv  引数                                                         
//!  @return           常に０                                                       
// -------------------------------------------------------------------------------- 
int RAMFUNC main(int argc, char *argv[])
{
	static char		state[]={ '-', 'O' } ;
	char			buff[32] ;
	int				index ;


	setup_default_uart() ;

	memset(&m_vram, 0, sizeof( m_vram )) ;
	m_spritePos = 0 ;
	m_bitmapPos = 0 ;

	// サインデータ作成
	for(index=0; index<256; index++) {
		m_sinData[index] = (int16_t )(sin( M_PI / (256 / 2) * index ) * FP_ONE) ;
	}

	// 表示データ作成
	setBitmap16() ;
	setSpriteData() ;
	setTextData() ;

	sound_start() ;

	vga_init() ;
	pad_init() ;
	vga_setBgBitmap320x240x16( (const uint32_t *)&m_vram.bg.bmp16[0][0] ) ;
	vga_setFgSprite(&m_vram.fg.spr[0], &g_sprPatData[0][0], &g_palData[0][0]) ;
	vga_setHSyncProc( hsync ) ;
	vga_start() ;

	// VSync 割り込み
	gpio_set_irq_enabled_with_callback(VGA_CFG_PIN_VSYNC, GPIO_IRQ_EDGE_FALL, true, isr_vsync) ;

	printf("\r\r\r\r") ;

	while( 1 ) {
		vga_waitVSync() ;

		// パッド ID
		sprintf(&buff[0], "ID=%d", pad_getId()) ;
		writeString(20, 0, &buff[0], 0) ;
		printf("%s / ", &buff[0]) ;

		// パッド状態
		sprintf(&buff[0], "%2d:", g_padInfo.type) ;
		for(index=0; index<16; index++) {
			buff[3+index] = 0xed - (((g_padInfo.status << index) >> 15) & 1) ;
		}
		buff[3+index] = '\0' ;
		writeString(20, 1, &buff[0], 0) ;

		for(index=0; index<16; index++) {
			buff[3+index] = state[(((g_padInfo.status<<index)>>15)&1)] ;
		}
		printf("%s / ", &buff[0]) ;

		sprintf(&buff[0], "0:%3d,%3d,1:%3d,%3d", g_padInfo.analog[0], g_padInfo.analog[1], g_padInfo.analog[2], g_padInfo.analog[3]) ;
		writeString(20, 2, &buff[0], 0) ;
		printf("%s  \r", &buff[0]) ;
	}

	return 0 ;
}
