// -------------------------------------------------------------------------------- 
//!  @file  sound.c                                                                 
//!  サウンド管理のソースファイル                                                   
//                                                                                  
//   Copyright (C) 2021 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <pico/stdlib.h>
#include "config.h"
#include "libvga/dac/mcp49x1.h"
#include "libvga/pcm/sn76489.h"
#include "libvga/pcm/scc.h"
#include "libvga/pcm/vrc6.h"
#include "libvga/pcm/namcot163.h"
#include "data.h"
#include "sound.h"


#ifdef  SOUNDDAC_CFG_DAC_MCP49x1
#define DAC_INIT					mcp49x1_init					//!< DAC 初期化処理
#define DAC_WRITE					mcp49x1_write					//!< DAC 書き込み処理
#define DAC_WAIT					mcp49x1_wait					//!< DAC 書き込み待ち処理
#endif//SOUNDDAC_CFG_DAC_MCP49x1


static struct repeating_timer		m_timer ;						//!< 繰り返しタイマー
static PCMSN76489INFO				m_sn76489Info ;					//!< SN76489 情報
static PCMSCCINFO					m_sccInfo ;						//!< SCC 情報
static PCMVRC6INFO					m_vrc6Info ;					//!< VRC6 情報
static PCMNAMCOT163INFO				m_namcot163Info ;				//!< NAMCOT-136 情報
static uint16_t						m_output ;						//!< 出力値
static uint8_t						m_type ;						//!< 種類
static uint16_t						const *m_p_data ;				//!< 再生データ


// -------------------------------------------------------------------------------- 
//!  繰り返しタイマー割り込みサービスルーチン                                       
//!                                                                                 
//!  @param [in] p_timer  繰り返しタイマー                                          
//!  @retval     true     繰り返し                                                  
//!  @retval     false    停止                                                      
// -------------------------------------------------------------------------------- 
static bool RAMFUNC isr_timer( struct repeating_timer *p_timer )
{
	int32_t		output ;


	DAC_WRITE( m_output ) ;

	output   = pcm_sn76489_update( &m_sn76489Info ) ;
	output  += pcm_scc_update( &m_sccInfo ) ;
	output  += pcm_vrc6_update( &m_vrc6Info ) ;
	output  += pcm_namcot163_update( &m_namcot163Info ) ;
	m_output = (uint16_t )(output + 0x0800) ;

	DAC_WAIT() ;

	return true ;
}


// -------------------------------------------------------------------------------- 
//!  動作開始                                                                       
//!                                                                                 
//!  @return  なし                                                                  
// -------------------------------------------------------------------------------- 
void sound_start( void )
{
	int		index ;


	DAC_INIT() ;

	m_output = 0x0800 ;
	m_type   = 0 ;
	m_p_data = &g_soundDataSN76489[0][0] ;

	pcm_sn76489_reset( &m_sn76489Info ) ;
	pcm_scc_reset( &m_sccInfo ) ;
	pcm_vrc6_reset( &m_vrc6Info ) ;
	pcm_namcot163_reset( &m_namcot163Info ) ;

	for(index=0; index<=0x0800; index+=0x0040) {
		DAC_WRITE( index ) ;
		DAC_WAIT() ;
		sleep_us( 10 ) ;
	}

	add_repeating_timer_us(-SOUNDDAC_CFG_TIME, isr_timer, NULL, &m_timer) ;
}


// -------------------------------------------------------------------------------- 
//!  VSync 処理                                                                     
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
void RAMFUNC sound_vsync( void )
{
	switch( m_type ) {
		case 0:		// SN76489
		{
			if( !*m_p_data ) {
				m_type = 1 ;

				pcm_sn76489_reset( &m_sn76489Info ) ;

				m_p_data = (uint16_t const *)&g_soundDataNamcot163[0] ;

				pcm_namcot163_copyWaveData(&m_namcot163Info, 0, (uint8_t const *)&m_p_data[0], 64) ;
				m_p_data += 16 ;

				pcm_namcot163_setChannelCount(&m_namcot163Info, 8) ;
				pcm_namcot163_setAddress(&m_namcot163Info, 0, 32, 56) ;
				pcm_namcot163_setAddress(&m_namcot163Info, 1, 32, 56) ;
				pcm_namcot163_setAddress(&m_namcot163Info, 2, 32, 56) ;
				pcm_namcot163_setAddress(&m_namcot163Info, 3, 32, 56) ;
				pcm_namcot163_setAddress(&m_namcot163Info, 4,  0, 60) ;
				pcm_namcot163_setAddress(&m_namcot163Info, 5,  0, 60) ;
				pcm_namcot163_setAddress(&m_namcot163Info, 6,  0, 60) ;
				pcm_namcot163_setAddress(&m_namcot163Info, 7,  0, 60) ;

				pcm_namcot163_enable(&m_namcot163Info, true) ;
				break ;
			}

			pcm_sn76489_setReg(&m_sn76489Info, 0, m_p_data[0]) ;
			pcm_sn76489_setReg(&m_sn76489Info, 1, m_p_data[1]) ;
			pcm_sn76489_setReg(&m_sn76489Info, 2, m_p_data[2]) ;
			pcm_sn76489_setReg(&m_sn76489Info, 3, m_p_data[3]) ;
			m_p_data += 4 ;
			break ;
		}

		case 1:		// NAMCOT-163
		{
			if( *(uint32_t *)&m_p_data[0] == 0xffffffff ) {
				m_type = 2 ;

				pcm_namcot163_reset( &m_namcot163Info ) ;

				pcm_scc_enable(&m_sccInfo, 0x00) ;
				pcm_scc_copyWaveData(&m_sccInfo, 0, (int8_t const *)&g_soundDataSCC[0], 128) ;
				pcm_scc_enable(&m_sccInfo, 0x1f) ;

				m_p_data = &g_soundDataSCC[128/2] ;
				break ;
			}

			m_namcot163Info.channelInfo[0].reg[0] = *(uint32_t *)&m_p_data[0] ;
			m_namcot163Info.channelInfo[1].reg[0] = *(uint32_t *)&m_p_data[2] ;
			m_namcot163Info.channelInfo[2].reg[0] = *(uint32_t *)&m_p_data[4] ;
			m_namcot163Info.channelInfo[3].reg[0] = *(uint32_t *)&m_p_data[6] ;
			m_namcot163Info.channelInfo[4].reg[0] = *(uint32_t *)&m_p_data[8] ;
			m_namcot163Info.channelInfo[5].reg[0] = *(uint32_t *)&m_p_data[10] ;
			m_namcot163Info.channelInfo[6].reg[0] = *(uint32_t *)&m_p_data[12] ;
			m_namcot163Info.channelInfo[7].reg[0] = *(uint32_t *)&m_p_data[14] ;
			m_p_data                             += 16 ;
			break ;
		}

		case 2:		// SCC
		{
			if( *m_p_data == 0xffff ) {
				m_type = 3 ;

				pcm_scc_reset( &m_sccInfo ) ;

				m_p_data = (uint16_t const *)&g_soundDataVRC6[0][0] ;
				break ;
			}

			pcm_scc_setReg(&m_sccInfo, 0, m_p_data[0]) ;
			pcm_scc_setReg(&m_sccInfo, 1, m_p_data[1]) ;
			pcm_scc_setReg(&m_sccInfo, 2, m_p_data[2]) ;
			pcm_scc_setReg(&m_sccInfo, 3, m_p_data[3]) ;
			pcm_scc_setReg(&m_sccInfo, 4, m_p_data[4]) ;
			m_p_data += 5 ;
			break ;
		}

		case 3:		// VRC6
		{
			if( *(uint32_t *)m_p_data == 0xffffffff ) {
				m_type = 0 ;

				pcm_vrc6_reset( &m_vrc6Info ) ;

				m_p_data = &g_soundDataSN76489[0][0] ;
				break ;
			}

			pcm_vrc6_setReg(&m_vrc6Info, 0, *(uint32_t *)&m_p_data[0]) ;
			pcm_vrc6_setReg(&m_vrc6Info, 1, *(uint32_t *)&m_p_data[2]) ;
			pcm_vrc6_setReg(&m_vrc6Info, 2, *(uint32_t *)&m_p_data[4]) ;
			m_p_data += 6 ;
			break ;
		}

		default:
			break ;
	}
}
